// VinciaShower.h is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2010 W.T. Giele, D.A. Kosower, P.Z. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// This file contains header information for the Antenna class 
  
#ifndef VINCIA_VinciaAntenna_H
#define VINCIA_VinciaAntenna_H

#include "Basics.h"
#include "Event.h"
#include "PythiaStdlib.h"
#include "VinciaCommon.h"

namespace Vincia {
  
  // Generic classes contained in this header
  class Antenna;
  class AntennaSet;

  // Derived classes, inheriting from the Antenna class
  // (more can be added, to define other types of antennae)
  // Each AntennaSet contains a list of Antennae.
  class GGEmit; 
  class QQEmit;
  class QGEmit;
  class GGSplit;
  class QGSplit;

  // Auxiliary classes
  class ThreeInt;
  class FourInt;
  class SixInt;
  
  //**************************************************************************

  // The Antenna class: 
  // Base antenna class which contains all the shared pieces and 
  // a specialization for each antenna type which overrides some 
  // of the virtual functions. 
  // The base class encodes a simple eikonal gluon-emission factor
  
  class Antenna {
    
  public:
    
    //Bare constructor: switch defaults on, nothing initialized yet.
  Antenna() : octetPartitioning(0.), deadZoneAvoidance(0.), verbose(1) {
      useFiniteTerms = true;
      useCollinearTerms = true;
      useMassTerms = true; 
      isInitPtr = false; 
      isOnSav = false;
    };
    
    // Destructor 
    virtual ~Antenna() {};
    
    // Initialize pointers:
    //    ParticleData : required for masses, color charges, etc
    //    Rndm : required to generate random test points to check antenna
    //    Settings : required for all coefficients, switches, etc.
    void initPtr(ParticleData* particleDataPtrIn, Rndm* rndmPtrIn, 
		 Settings* settingsPtrIn) {
      particleDataPtr = particleDataPtrIn; 
      rndmPtr         = rndmPtrIn;
      settingsPtr     = settingsPtrIn; 
      isInitPtr       = true;
    };

    // Default initialization should be good enough for most antennae.
    // Leave it virtual so future implementations can in principle extend it
    virtual bool init(bool checkAntenna=true);

    // Antenna function name(s) 
    //    name : technical name, used to identify parameters and switches 
    //           in XML, cmnd files etc.
    //    humanName : for info only, can be whatever you want
    virtual string name()    const {return "Vincia:IK:Emit";}
    virtual string humanName()    const {return "g/IK";}

    // Function to encode singular and non-singular antenna function pieces 
    virtual double antennaFunction(double yij, double yjk, double yik, 
			  double mu1 = 0., double mu2 = 0., double mu3 = 0.);      

    // Functions to get and set antenna function coefficients
    virtual void readC();              // Read coefficients from settings database
    virtual bool setC(int i=0,int j=0,double c=0.0); // Set C[i][j]
    virtual double getC(int i, int j);
    virtual void clearC();             // Clear all coefficients
    
    // MR 16.7.2010: list the coefficients of the antenna
    virtual void listCoefficients(ostream& os=cout);
    
    // rescale finite terms
    virtual bool rescaleFiniteTerms(double rescalingFactor);
               
    // get test masses used in Initialization
    virtual vector< vector<double> > gettestMasses();
    
    // END OF VIRTUAL FUNCTIONS

    // Alternative version of init that includes pointers and automatically
    // calls initPtr, so only one complete initialization call needed.
    bool init(bool checkAntenna, ParticleData* particleDataPtrIn, 
	      Rndm* rndmPtrIn, Settings* settingsPtrIn) {
      initPtr(particleDataPtrIn, rndmPtrIn, settingsPtrIn);
      return init(checkAntenna);
    }

    // Eikonal function, used as convenient shorthand in gluon-emission antennae
    double aEikonalY(double yij, double yjk, double yik,
		    double mui=0., double muk=0.);    

    // Shorthand for sum of non-singular coefficients, the F functions in [GRS11].
    double aTaylorY(double yij, double yjk);    
    
    // Non-virtual functions to get the value of the antenna functions
    //   aBarY = antennaFunction     (dimensionless)
    //   aBar  = 1/sIK aBarY         (dimensionful)
    // Derived classes should modify the aSingY and aFiniteY pieces below.
    double aBarY(double yij, double yjk, 
		 double mui = 0., double muj = 0., double muk = 0.,
		 bool forcePositive = true);
    double aBar(double sij, double sjk, double s, 
		double massi = 0., double massj = 0., double massk = 0.,
		bool forcePositive = true);
    
    // Functions to set antenna parameters
    bool setParm(bool&, string);
    bool setParm(int&, string);
    bool setParm(double&, string);
    bool setParm(string&, string);
    
    // Functions to get antenna parameters
    bool   isOn() {return (isInit ? isOnSav : false);}
    int    getEvolutionType() { return evolutionType;}     
    int    getKineMapType()   { return kineMapType;  }  
    double getChargeFactor()  { return chargeFactor; }  
    
    // check combines IsPositive and IsFreeOfDeadZones
    bool check(int nPointsCheck, double mu1, double mu2, 
	       double mu3, bool dorepair=false);

    // MR 21.12.2010: IsPositive checks whether an antenna is positive
    // given a set of masses. It has an optional argument dorepair 
    // (default false) which governs whether the finite coefficients 
    // are changed to make the antenna positive.
    bool IsPositive( int nPointsCheck, double mu1, double mu2, 
		     double mu3, bool dorepair = false );

    // MR 21.12.2010: IsFreeOfDeadZones checks whether an antenna has 
    // dead zones given a set of masses. It has an optional argument 
    // dorepair (default false) which governs whether the finite 
    // coefficients are changed to make the dead zones go away.
    bool IsFreeOfDeadZones( 
      int nPointsCheck, double mu1, double mu2, double mu3, 
      bool dorepair = false );

    // check function called by init
    bool checkInit(vector< vector<double> > testMasses);
      
  protected:
    
    // Antenna Parameters 
    int evolutionType;
    int kineMapType;
    double chargeFactor, octetPartitioning;
    
    // In- and Out- particles in PDG code format
    vector<int> idOld;
    vector<int> idNew;
    vector<string> nameOld;
    vector<string> nameNew;
    
    // Coefficients of the antenna function 
    map<int, map<int, double> > LaurentC;
    
    // Flag whether to use finite, collinear, and mass terms or not
    bool useFiniteTerms, useCollinearTerms, useMassTerms;

    // Dead Zone Avoidance Factor
    double deadZoneAvoidance;
    
    // Initialized flag
    bool isInit;

    // On/off flag
    bool isOnSav;
    
    // Pointers to Pythia-8 classes
    bool isInitPtr;
    ParticleData* particleDataPtr;
    Rndm*         rndmPtr;
    Settings*     settingsPtr;  
    
    // Verbose setting
    int verbose;
  };
  
  //**************************************************************************
  // The QQEmit class: 
  // Inherits from Antenna
  // Contains the antenna function for (3,-3) -> (3,8,-3)

  class QQEmit : public Antenna {

  public:

    // constructor
  QQEmit() : Antenna() {};

    // Name
    virtual string name()    const {return "Vincia:qq:Emit";}
    virtual string humanName()    const {return "g/qq";}

    // Antenna function
    virtual double antennaFunction(double yij, double yjk, double yik, 
			   double mu1=0., double mu2=0., double mu3=0.);
    
    // Other inherited functions
    virtual void readC();
    virtual void clearC();      
    virtual void listCoefficients(ostream& os=cout);      
    virtual bool rescaleFiniteTerms(double rescalingFactor);      
    virtual vector< vector<double> > gettestMasses();

    // Additional member functions
    double getM(int mu1ord, int mu3ord, int y12ord, int y23ord);
    bool setM(int mu1ord, int mu3ord, int y12ord, int y23ord, double mIn);
      
    double getDenominatorMassPrefactor();
    bool setDenominatorMassPrefactor(double pfIn);
    
  private:
    
    double denominatorMassPrefactor;
    map<FourInt, double> LaurentM;
  };
  
  //**************************************************************************
  // The QGEmit class: 
  // Inherits from Antenna
  // Contains the antenna function for (3,8) -> (3,8,8)

  class QGEmit : public Antenna {

  public:

    // constructor
  QGEmit() : Antenna() {};

    // Name
    virtual string name()    const {return "Vincia:qg:Emit";}
    virtual string humanName()    const {return "g/qg";}

    // Antenna function
    double antennaFunction(double yij, double yjk, double yik, 
			   double mu1=0., double mu2=0., double mu3=0.);
        
    // Other inherited functions 
    virtual void readC();
    virtual void clearC();
    virtual void listCoefficients(ostream& os=cout);      
    virtual bool rescaleFiniteTerms(double rescalingFactor);      
    virtual vector< vector<double> > gettestMasses();

    // Additional member functions
    double getM(int muord, int y12ord, int y23ord);
    bool setM(int muord, int y12ord, int y23ord, double mIn);
      
    double getMtilde(int mu1ord, int y12ord, int y13ord, int y23ord);
    bool setMtilde(
		   int mu1ord, int y12ord, int y13ord, int y23ord, double mIn);
    
  private:
    
    map<ThreeInt, double> LaurentM;
    map<FourInt, double> LaurentMtilde;
  };
  
  //**************************************************************************
  // The GGEmit class: 
  // Inherits from Antenna
  // Contains the antenna function for (8,8) -> (8,8,8)

  class GGEmit : public Antenna {

  public:
    
    // constructor
  GGEmit() : Antenna() {};

    // Name
    virtual string name()    const {return "Vincia:gg:Emit";}
    virtual string humanName()    const {return "g/gg";}

    // Antenna function
    virtual double antennaFunction(double yij, double yjk, double yik, 
			   double mu1=0.,double mu2=0., double mu3=0.);    

  private:
    
  };
  
  //**************************************************************************
  // The QGSplit class:
  // Inherits from Antenna
  // Contains the antenna function for (3,8) -> (3,-3,3)

  class QGSplit : public Antenna {

  public:

    // constructor
  QGSplit() : Antenna() {};

    // Name
    virtual string name()    const {return "Vincia:qg:Split";}
    virtual string humanName()    const {return "q'/qg";}

    // Antenna function
    virtual double antennaFunction(double yij, double yjk, double yik, 
				   double mu1=0., double mu2=0., double mu3=0.);
    

    // Other inherited functions 
    virtual void readC();
    virtual void clearC();
    virtual void listCoefficients(ostream& os=cout);      
    virtual bool rescaleFiniteTerms(double rescalingFactor);      
    virtual vector< vector<double> > gettestMasses();

    // Additional member functions
    double getM(int muqord, int muqpord, int y12ord, int y23ord);
    bool setM(
	      int muqord, int muqpord, int y12ord, int y23ord, double mIn);
    
    double getMtilde(int muqord, int muqpord, int y12ord, int y23ord);
    bool setMtilde(
		   int muqord, int muqpord, int y12ord, int y23ord, double mIn);
    
    double getCtilde(int y12ord, int y23ord);
    bool setCtilde(int y12ord, int y23ord, double cIn);
    
  private:
    
    map<FourInt, double> LaurentM;
    map<FourInt, double> LaurentMtilde;
    map< pair<int, int>, double > LaurentCtilde;
  };
  
  //**************************************************************************
  // The GGSplit class:
  // Inherits from Antenna
  // Contains the antenna function for (8,8) -> (8,-3,3)

  class GGSplit : public Antenna {
    
  public:
    
    // constructor
  GGSplit() : Antenna() {};
    
    // Name
    virtual string name()    const {return "Vincia:gg:Split";}
    virtual string humanName()    const {return "q/gg";}

    // Antenna function
    virtual double antennaFunction(double yij, double yjk, double yik, 
			   double mu1=0.,double mu2=0., double mu3=0.);

    // Other inherited functions 
    virtual void readC();
    virtual void clearC();
    virtual void listCoefficients(ostream& os=cout);      
    virtual bool rescaleFiniteTerms(double rescalingFactor);      
    virtual vector< vector<double> > gettestMasses();

    // Additional member functions
    double getM(int muqpord, int y12ord, int y23ord);
    bool setM(
	      int muqpord, int y12ord, int y23ord, double mIn);
    
    double getCtilde(int y12ord, int y23ord);
    bool setCtilde(int y12ord, int y23ord, double cIn);    

  private:    
    
    map<ThreeInt, double> LaurentM;
    map< pair<int, int>, double> LaurentCtilde;
  };
  
  //**************************************************************************
  // The AntennaSet class:
  // Contains a list of Antennae (e.g., the "DEF" set, the "GGG" set, etc)

  class AntennaSet {

  public:
    // destructor
    // delete the antennae
    ~AntennaSet(){
      for (
        vector<Antenna*>::iterator antPtIt = antPtrs.begin(); 
        antPtIt != antPtrs.end(); ++antPtIt
        ) {
        delete *antPtIt;
      }
    }
    // initialisation
    void init(int verbose, bool checkAntennae=false) {
      if (! isInitPtr) {
        printOut("AntennaSet::init",
                  "Cannot initialize, pointers not set.");
        return;
      }
      // Read in and initialize generic Sudakovs
      antPtrs.resize(5);
      antPtrs[iqqEmit] = new QQEmit();
      antPtrs[iqgEmit] = new QGEmit();
      antPtrs[iggEmit] = new GGEmit();
      antPtrs[iqgSplit] = new QGSplit();
      antPtrs[iggSplit] = new GGSplit();
      // loop through antPtrs and initialise
      for (
        vector<Antenna*>::iterator antPtIt = antPtrs.begin(); 
        antPtIt != antPtrs.end(); ++antPtIt
        ) {
        // initialise pointers for antennae, giving pointers as args
        bool pass = 
          (*antPtIt)->init(
            checkAntennae, particleDataPtr, rndmPtr, settingsPtr);
        if (pass) {	
          // Verbose output
          if (verbose >= 3)
            printOut((*antPtIt)->humanName()+":init",
                                          "ok: added to antenna list");
        } 
        else if (verbose >= 1) {
            printErr((*antPtIt)->name()+":init",
              (*antPtIt)->name()+
                  " has a problem and will not be used in this run!");
            *antPtIt = NULL;
        }
      }
    };
    void initPtr(ParticleData* particleDataPtrIn, Rndm* rndmPtrIn, 
		 Settings* settingsPtrIn) {
      particleDataPtr = particleDataPtrIn; 
      rndmPtr         = rndmPtrIn;
      settingsPtr     = settingsPtrIn; 
      isInitPtr       = true;
    }
    
    // at present this does only allow to set generic Antennae,
    // Antennae with a derived type will be converted to Antenna
    void setAnt(vector<Antenna*>::size_type iAntIn, Antenna antIn) {
      if (iAntIn > antPtrs.size()) {
        antPtrs.resize(iAntIn);
      }
      Antenna* antCopyPtr = new Antenna(antIn);
      antPtrs[iAntIn] = antCopyPtr;
      antIndxSav[antCopyPtr->name()]=iAntIn;
    }
    
    bool exists(int iAntIn) {
      return (iAntIn<0 || iAntIn >= int(antPtrs.size())) ? false : true;
    }
    Antenna* getAnt(vector<Antenna*>::size_type iAntIn) {
      if (iAntIn <= antPtrs.size()) {
        return antPtrs[iAntIn];
      }
      else return NULL;
    }
    string name(vector<Antenna*>::size_type iAntIn) {
      if (iAntIn <= antPtrs.size()) {
        return antPtrs[iAntIn]->name();
      }
      else return "none";
    }
    string humanName(vector<Antenna*>::size_type iAntIn) {
       if (iAntIn <= antPtrs.size()) {
        return antPtrs[iAntIn]->humanName();
      }
      else return "none";
    }
    int getIndx(string antNameIn) {return antIndxSav[antNameIn];}
    
  private:
    // use a vector of Antenna pointers, create them with new on 
    // initialization
    vector<Antenna*> antPtrs;    
    map<string,int> antIndxSav;
    // Pointers to Pythia-8 classes, needed to initialise antennae
    bool isInitPtr;
    ParticleData* particleDataPtr;
    Rndm*         rndmPtr;
    Settings*     settingsPtr;  
  }; // End class AntennaSet
  
  // the nInt classes are needed as indices for STL maps
  // should be a template, but the code is not that long
  class ThreeInt {
  public:
    // Constructor
  ThreeInt(int i1, int i2, int i3) :
    int1(i1), int2(i2), int3(i3) {}
    // comparison
    bool operator<(const ThreeInt &right) const {
      return (
	      int1 < right.int1 ||
	      (int1 == right.int1 && int2 < right.int2) ||
	      (int1 == right.int1 && int2 == right.int2 &&
	       int3 < right.int3) 
	      );
    }
    int int1;
    int int2;
    int int3;
  }; // End class ThreeInt
  
  class FourInt {
  public:
    // Constructor
  FourInt(int i1, int i2, int i3, int i4) :
    int1(i1), int2(i2), int3(i3), int4(i4) {}
    // comparison
    bool operator<(const FourInt &right) const {
      return (
	      int1 < right.int1 ||
	      (int1 == right.int1 && int2 < right.int2) ||
	      (int1 == right.int1 && int2 == right.int2 &&
	       int3 < right.int3) ||
	      (int1 == right.int1 && int2 == right.int2 &&
	       int3 == right.int3 && int4 < right.int4)
	      );
    }
    int int1;
    int int2;
    int int3;
    int int4;
  }; // End class FourInt
  
  class SixInt {
  public:
    // Constructor
  SixInt(int i1, int i2, int i3, int i4, int i5, int i6) :
    int1(i1), int2(i2), int3(i3), int4(i4), int5(i5), int6(i6) {}
    // Return operator
    bool operator<(const SixInt &right) const {
      return (
	      int1 < right.int1 ||
	      (int1 == right.int1 && int2 < right.int2) ||
	      (int1 == right.int1 && int2 == right.int2 &&
	       int3 < right.int3) ||
	      (int1 == right.int1 && int2 == right.int2 &&
	       int3 == right.int3 && int4 < right.int4) ||
	      (int1 == right.int1 && int2 == right.int2 &&
	       int3 == right.int3 && int4 == right.int4 &&
	       int5 < right.int5) ||
	      (int1 == right.int1 && int2 == right.int2 &&
	       int3 == right.int3 && int4 == right.int4 &&
	       int5 == right.int5 && int6 < right.int6)
	      
	      );
    }
    // maybe this should be private, not sure
    int int1;
    int int2;
    int int3;
    int int4;
    int int5;
    int int6;
  }; // End class SixInt
  
} // End Vincia namespace

#endif
