// vincia05-root is a part of the VINCIA plugin to PYTHIA 8
// This is a simple test program to run Z->bbbar at LEP 1, 
// interfaced to ROOT, including a runtime updated display,
// using VINCIA's ROOT-based runtime display. 
// Uses VINCIA for showering and matching and PYTHIA 8 for all else
// Copyright (c) 2011 Peter Skands

// PYTHIA and VINCIA include files
#include "Vincia.h"
// ROOT: include files for linking to ROOT, uncomment as necessary
#include "VinciaRoot.h"
// FASTJET
// #include "fastjet/PseudoJet.hh"
// #include "fastjet/ClusterSequence.hh"

// Use Pythia8 and Vincia namespaces, for convenience.
using namespace Pythia8; 
using namespace Vincia; 

// Main Program
int main() {

  //************************************************************************
  // PYTHIA and VINCIA objects
  //************************************************************************

  // Create Pythia and Vincia objects
  Pythia pythia; 
  VinciaPlugin vincia(&pythia,"vincia05.cmnd");

  //************************************************************************
  // Define some useful pointers and shorthands for later on
  //************************************************************************

  // Shorthand for event record
  Event& event = pythia.event;
  Settings* settingsPtr = vincia.settingsPtr;        // to settings database

  // Extract settings to be used in the main program.  
  double eCM       = settingsPtr->parm("Beams:eCM");
  int    nEvent    = settingsPtr->mode("Main:numberOfEvents");
  int    nShow     = settingsPtr->mode("Main:timesToShow");
  bool   vinciaOn  = settingsPtr->flag("Vincia");
  bool   hadronize = settingsPtr->flag("HadronLevel:all");
  int    verbose   = settingsPtr->mode("Vincia:verbose");

  // Extrac matching settings
  int    maxMatchTree     = settingsPtr->mode("VinciaMatching:order");
  bool   matching         = maxMatchTree >= 1;

  //************************************************************************
  // VinciaRoot
  //************************************************************************

  // Create VinciaRoot objects
  VinciaRoot vroot; 

  // Initialize some ROOT displays (name,x-pixels,y-pixels,x-offset,y-offset)
  // Book additional displays here as you like.
  vroot.createDisplay("Thrust",300,400);
  vroot.createDisplay("C",300,400);
  vroot.createDisplay("D",300,400);
  vroot.createDisplay("MH",300,400);
  vroot.createDisplay("BW",300,400);
  vroot.createDisplay("BT",300,400);
  vroot.createDisplay("NCh",300,400);
  vroot.createDisplay("xCh",300,400);
  vroot.createDisplay("xB",300,400);

  // Tell VinciaRoot which verbose level to use
  vroot.setVerbose(verbose);
  // Switch stats on/off (in ROOT conventions, but with 0 = off)
  vroot.setOptStat(0);
  // Tell it the name of the MC we are running (for plot legends)
  string pyVer = num2str(settingsPtr->parm("Pythia:versionNumber"),5);  
  string vcVer = num2str(settingsPtr->parm("Vincia:versionNumber"),5);  
  if (vinciaOn && matching) {
    vroot.setLabel("Vincia");    
    vroot.setRefs("Vincia "+vcVer+" + MadGraph 4.426 + Pythia "+pyVer);
  } else if (vinciaOn) {
    vroot.setLabel("Vincia");
    vroot.setRefs("Vincia "+vcVer+" + Pythia "+pyVer);
  } else {
    vroot.setLabel("Pythia 8");
    vroot.setRefs("Pythia "+pyVer);
  }

  // Use a different color scheme for PYTHIA plots
  if ( !vinciaOn ) vroot.setColorScheme(2);

  //************************************************************************
  // Initialize PYTHIA (automatically initializes VINCIA) 
  //************************************************************************

  // Initialize
  pythia.init( 11, -11, eCM);

  // List changed settings
  pythia.settings.listChanged();

  // Check that Z0 decay channels set correctly.
  pythia.particleData.listChanged();
 
  //***************************************************************************
  // ROOT
  //***************************************************************************

  // Form name for ROOT file
  ostringstream outfile;
  // Put root files in subdirectory defined at beginning of file
  if (not vinciaOn) {
    outfile<<"pythia";
  } else {
    outfile<<"vincia";
  }
  if (settingsPtr->flag("HadronLevel:all")) outfile<<"-had";
  outfile<<".root";

  //***************************************************************************
  // Book Histograms

  // Event Weights
  TH1F *HnPN=new TH1F("Weights","Event Weights", 200, -2.5, 2.5);
  //  vroot.addTH1FPtr(HnPN,"MyPlots",true,false,false);
  HnPN->SetFillColor(3); 

  // Quark Pairs
  TH1F *HnQuarks=new TH1F("nQuarks","Number of Quark Pairs", 15, -0.5, 14.5);
  //vroot.addTH1FPtr(HnQuarks,"MyPlots",true,true,true);
  HnQuarks->SetMinimum(0.0001);
  HnQuarks->SetMaximum(0.75);

  // nPartons
  TH1F *HnPartons=new TH1F("nPartons","Number of Partons", 60, -0.5, 59.5);
  // vroot.addTH1FPtr(HnPartons,"MyPlots",true,true,true);
  HnPartons->SetMinimum(0.001);
  HnPartons->SetMaximum(0.25);

  // Thrust
  TH1F *HThrust=new TH1F("Thrust","1-T (b)", 100, 0.0, 0.4);
  double wThrust = 100.0/0.4;
  HThrust->GetXaxis()->SetTitle("1-T (b)");
  HThrust->GetYaxis()->SetTitle("1/N dN/d(1-T)");
  HThrust->SetMinimum(0.0005);
  HThrust->SetMaximum(100.);
  vroot.addTH1FPtr(HThrust,"Thrust",true,true,true,"data/L3-91-Thrust-b.vec","py-T-b.dat");

  // C
  TH1F *HC=new TH1F("C","C parameter (b)", 40, 0.0, 1.0);
  double wC = 40.0/1.0;
  HC->GetXaxis()->SetTitle("C (b)");
  HC->GetYaxis()->SetTitle("1/N dN/dC");
  HC->SetMinimum(0.0005);
  HC->SetMaximum(100.);
  vroot.addTH1FPtr(HC,"C",true,true,true,"data/L3-91-C-b.vec","py-C-b.dat");

  // D (4-jet obs)
  TH1F *HD=new TH1F("D","D parameter (b)", 50, 0.0, 0.8);
  double wD = 50.0/0.8;
  HD->GetXaxis()->SetTitle("D (b)");
  HD->GetYaxis()->SetTitle("1/N dN/dD");
  HD->SetMinimum(0.0005);
  HD->SetMaximum(100.);
  vroot.addTH1FPtr(HD,"D",true,true,true,"data/L3-91-D-b.vec","py-D-b.dat");
  
  // Heavy-Jet Mass (3-jet obs)
  TH1F *HMjH=new TH1F("HeavyJetMass","Heavy Jet Mass (b)",40, 0.0, 0.3);
  double wMjH = 40.0/0.3;
  vroot.addTH1FPtr(HMjH,"MH",true,true,true,"data/L3-91-MH-b.vec","py-MH-b.dat");
  HMjH->SetMaximum(100.01); 
  HMjH->SetMinimum(0.005); 
  
  // Wide Jet Broadening
  TH1F *HBjWide=new TH1F("WideJetBroadening","Wide Jet Broadening (b)",40, 0.0, 0.3);
  double wBjWide = 40.0/0.3;
  HBjWide->GetXaxis()->SetTitle("B_{W} (b)");
  HBjWide->GetYaxis()->SetTitle("1/N dN/dB_{W}");
  HBjWide->SetMaximum(100.01); 
  HBjWide->SetMinimum(0.005);   
  vroot.addTH1FPtr(HBjWide,"BW",true,true,true,"data/L3-91-BWide-b.vec","py-BW-b.dat");

  // Total Jet Broadening
  TH1F *HBjTot=new TH1F("TotalJetBroadening","Total Jet Broadening (b)",40, 0.0, 0.4);
  double wBjTot = 40.0/0.4;
  HBjWide->GetXaxis()->SetTitle("B_{T} (b)");
  HBjWide->GetYaxis()->SetTitle("1/N dN/dB_{T}");
  HBjTot->SetMaximum(100.01); 
  HBjTot->SetMinimum(0.005);   
  vroot.addTH1FPtr(HBjTot,"BT",true,true,true,"data/L3-91-BTot-b.vec","py-BT-b.dat");

  // nCharged
  TH1F *HnCharged=new TH1F("nCharged","nCharged (b)", 60, -0.5, 59.5);
  HnCharged->GetXaxis()->SetTitle("N_{ch} (b)");
  HnCharged->GetYaxis()->SetTitle("Probability(N_{ch})");
  HnCharged->SetMinimum(0.00002);
  HnCharged->SetMaximum(0.5);
  vroot.addTH1FPtr(HnCharged,"NCh",true,true,true,"data/L3-91-Nch-b.vec");

  // Ln(1/x)
  TH1F *Hx=new TH1F("LnXInv","Ln(1/x) (b)", 200, 0.0, 8.0);
  double wx = 200.0/8.0;
  Hx->GetXaxis()->SetTitle("Ln(1/x) (b)");
  Hx->GetYaxis()->SetTitle("1/N_{ch} dN_{ch}/dLn(1/x)");
  Hx->SetMinimum(0.0002);
  Hx->SetMaximum(5.);
  vroot.addTH1FPtr(Hx,"xCh",true,true,true,"data/L3-91-LogInvx-b.vec");

  // xB
  TH1F *HxB=new TH1F("XB","xB", 50, 0.0, 1.0);
  double wxB = 50.0/1.0;
  HxB->GetXaxis()->SetTitle("x_{B}");
  HxB->GetYaxis()->SetTitle("1/N dN/dx_{B}");
  HxB->SetMinimum(0.02);
  HxB->SetMaximum(20.);
  vroot.addTH1FPtr(HxB,"xB",true,true,true,"data/SLD-91-xBweak.vec","py-xB.dat");

  // Event analysis tools
  Thrust Thrust(1);
  Sphericity SphLin(1,2),SphSq(2,2);
  ClusterJet ktClu("Durham",2,2);

  //***************************************************************************
  // EVENT GENERATION LOOP
  // Generation, event-by-event printout, analysis, and storage
  //***************************************************************************

  // (Optionally) re-initialize random number engine before event loop
  const int initseed = settingsPtr->mode("Random:seed");
  vincia.rndmPtr->init(initseed);

  // Counter for negative-weight events
  double nNegative = 0.0;
  double nPositive = 0.0;
  double evwt=1.0;

  // Number of quarks, partons, and charged particles
  double nQsum=0.0;
  double nPsum=0.0;
  double nCsum=0.0;

  // Begin event loop
  for (int iEvent = 0; iEvent < nEvent; ++iEvent) {

    if (iEvent%(max(1,nEvent/max(nShow,1))) == 0)  
      cout << " Now begin event " << iEvent << endl;

    if (!pythia.next()) continue;

    // Check for negative weights
    if (vinciaOn) evwt = vincia.weight();
    if (evwt < 0) nNegative += abs(evwt); 
    else nPositive += evwt;
    HnPN->Fill(evwt,1.0);

    // Count FS charged hadrons, partons, and quarks
    int nc=0;
    int np=0;
    int nq=0;

    // Compute visible energy
    Vec4 pVis(0.0);
    for (int i=1;i<event.size();i++) {
      // Final state only
      if (!pythia.event[i].isFinal())        continue;
      
      // No neutrinos
      if (pythia.event[i].idAbs() == 12 || pythia.event[i].idAbs() == 14 ||
          pythia.event[i].idAbs() == 16)     continue;
      
      pVis += event[i].p();
    }
    // Energy component
    double eVis = pVis.e();
						
    for (int i=1;i<event.size();i++) {
      // Count up final-state charged hadrons
      if (event[i].isCharged() && event[i].status() > 80) {
	nc++;
	double LnXInv=log(double(1.0/(2.0*event[i].pAbs()/eCM)));
	Hx->Fill(LnXInv,evwt*wx);
      }
      // Find last parton-level partons
      int iDau1 = event[i].daughter1();
      if (iDau1 == 0 || abs(event[iDau1].status()) > 80) {
	// Count up partons and quarks
	if (event[i].isQuark() || event[i].isGluon() ) np++;
	if (event[i].isQuark()) nq++;	
      }
      double x = 2.*event[i].pAbs() / eCM;

      // Find weakly decaying B mesons 
      int idp = abs(event[i].id());
      double xBweak = -1.0;
      if (hadronize) {
	if (idp > 510 && idp < 600) {
	  if (abs(event[event[i].daughter1()].id()) < 500) {
	    xBweak = x;
	  }
	}
	// Find weakly decaying B baryons
	if (idp > 5100 && idp < 6000) {
	  if (abs(event[event[i].daughter1()].id()) < 5000) {
	    xBweak = x;
	  }	
	}
      } 
      // If hadronization switched off, just plot x of b quarks
      else {
	if (idp == 5 && event[i].isFinal()) xBweak = x;
      }
      if (xBweak > 0.) HxB->Fill(xBweak,wxB);
	
    }

    // Fill Histograms
    nQsum += 0.5*nq;
    nPsum += np;
    nCsum += nc;
    HnQuarks->Fill(0.5*nq,evwt);
    HnPartons->Fill(1.0*np,evwt);    
    HnCharged->Fill(1.0*nc,evwt);

    // Event Shape analysis
    Thrust.analyze( event );
    HThrust->Fill(1-Thrust.thrust(),wThrust*evwt);

    // Histogram Linear Sphericity values
    if (np >= 2.0) {
      SphLin.analyze( event );
      SphSq.analyze( event );
      double evC=3*(SphLin.eigenValue(1)*SphLin.eigenValue(2)
		    + SphLin.eigenValue(2)*SphLin.eigenValue(3)
		    + SphLin.eigenValue(3)*SphLin.eigenValue(1));
      double evD=27*SphLin.eigenValue(1)*SphLin.eigenValue(2)
	*SphLin.eigenValue(3);
      HC->Fill(evC,wC*evwt);
      HD->Fill(evD,wD*evwt);
    }


    // Histogram jet masses
    Vec4 thrustAxis = Thrust.eventAxis(1);   
    Vec4 majorAxis = Thrust.eventAxis(2);   
    Vec4 minorAxis = Thrust.eventAxis(3);   
    Vec4 sphAxis = SphSq.eventAxis(1);   
    Vec4 sphAxis2 = SphSq.eventAxis(2);   
    Vec4 sphAxis3 = SphSq.eventAxis(3);   
    Vec4 pPlus, pMinus;
    double bPlus(0.), bMinus(0.), pPlusAbs(0.), pMinusAbs(0.), pTout(0.), pTin(0.);
    double pToutSph(0.), pTinSph(0.);
    // Loop over event to compute masses
    for (int i=0; i<event.size(); i++) {
      if (not event[i].isFinal() ) continue;
      int id = event[i].id();
      if (abs(id) == 12 || abs(id) == 14 || abs(id) == 16) continue;
      double dotp = event[i].px()*thrustAxis.px()
	+ event[i].py()*thrustAxis.py()
	+ event[i].pz()*thrustAxis.pz();
      double pAbs = event[i].pAbs();
      double cosT = dotp/pAbs/thrustAxis.pAbs();
      double sinT = sqrt(1.-pow2(cosT));
      if (dotp > 0) {
	pPlus     += event[i].p();
	bPlus     += abs(pAbs * sinT);
	pPlusAbs  += pAbs;
      } else {
	pMinus += event[i].p();
	bMinus    += abs(pAbs * sinT);
	pMinusAbs += pAbs;
      }
      if (event[i].isCharged()) {
	pTin  = abs(event[i].px()*majorAxis.px() + event[i].py()*majorAxis.py()
		    +event[i].pz()*majorAxis.pz());
	pTout = abs(event[i].px()*minorAxis.px() + event[i].py()*minorAxis.py()
		    +event[i].pz()*minorAxis.pz());
	pTinSph  = abs(event[i].px()*sphAxis2.px() + event[i].py()*sphAxis2.py()
		    +event[i].pz()*sphAxis2.pz());
	pToutSph = abs(event[i].px()*sphAxis3.px() + event[i].py()*sphAxis3.py()
		    +event[i].pz()*sphAxis3.pz());
      }
    }
    bPlus  /= 2.*(pPlusAbs+pMinusAbs);
    bMinus /= 2.*(pMinusAbs+pPlusAbs);    
    double mPlus  = pPlus.mCalc();
    double mMinus = pMinus.mCalc();
    HMjH->Fill(pow2(max(mPlus,mMinus)/eVis),wMjH*evwt);
    HBjWide->Fill(max(bPlus,bMinus),wBjWide*evwt);
    HBjTot->Fill(bPlus+bMinus,wBjTot*evwt);   
 
    // Periodically update displays
    if (iEvent == 9 || iEvent == 99 || iEvent%500 == 99|| iEvent == nEvent-1) {
      vroot.updateDisplay();       
      if ( (iEvent >= 1000 && iEvent%5000==99) || iEvent == nEvent-1) {
	string mcStr="vc";
	if (not vinciaOn) mcStr="py";
	else {
	  mcStr += "-m"+num2str(maxMatchTree,1);
	}
	vroot.saveDisplay("BT",mcStr,"eps");
	vroot.saveDisplay("BW",mcStr,"eps");
	vroot.saveDisplay("MH",mcStr,"eps");
	vroot.saveDisplay("D",mcStr,"eps");
	vroot.saveDisplay("C",mcStr,"eps");
	vroot.saveDisplay("Thrust",mcStr,"eps");
	vroot.saveDisplay("NCh",mcStr,"eps");
	vroot.saveDisplay("xCh",mcStr,"eps");
	vroot.saveDisplay("xB",mcStr,"eps");

	// Save Pythia histograms for overplotting later
	if (! vinciaOn) {
	  vroot.saveHistogram(HC,mcStr,"C-b","dat");
	  vroot.saveHistogram(HD,mcStr,"D-b","dat");
	  vroot.saveHistogram(Hx,mcStr,"x-b","dat");
	  vroot.saveHistogram(HxB,mcStr,"xB","dat");
	  vroot.saveHistogram(HnCharged,mcStr,"Nch-b","dat");
	  vroot.saveHistogram(HMjH,mcStr,"MH-b","dat");
	  vroot.saveHistogram(HBjWide,mcStr,"BW-b","dat");
	  vroot.saveHistogram(HBjTot,mcStr,"BT-b","dat");
	  vroot.saveHistogram(HThrust,mcStr,"T-b","dat");
	}
	
      }
    }

  }

  // Print out Vincia internal histograms
  if (vinciaOn && verbose >= 2) vincia.printHistos();

  // Print out Vincia end-of-run information
  pythia.statistics();
  if (vinciaOn) vincia.printInfo();

  // Save histograms in ROOT file
  cout<<"Saving root file... "<<outfile.str()<<endl;
  // Open ROOT file
  TFile f(outfile.str().c_str(),"recreate");
  HnPN->Write();
  HnQuarks->Write();
  HnPartons->Write();
  HnCharged->Write();
  Hx->Write();
  HThrust->Write();
  HC->Write();
  HD->Write();
  HMjH->Write();
  HBjWide->Write();
  HBjTot->Write();
  f.Close();

  cout << endl << "Enter q to terminate..." << endl;
  char tmp;cin>>tmp;
  
  return 0;
}
